#include <grp.h>
#include <luna/Units.h>
#include <os/ArgumentParser.h>
#include <os/Directory.h>
#include <os/File.h>
#include <os/FileSystem.h>
#include <pwd.h>

void find_user_and_group(struct stat& st, StringView& owner, StringView& group)
{
    auto* pw = getpwuid(st.st_uid);
    if (!pw) owner = "???";
    else
        owner = pw->pw_name;

    auto* grp = getgrgid(st.st_gid);
    if (!grp) group = "???";
    else
        group = grp->gr_name;
}

Result<int> luna_main(int argc, char** argv)
{
    StringView pathname;
    bool show_all { false };
    bool show_almost_all { false };
    bool long_list { false };
    bool human_readable { false };
    bool si { false };
    bool follow_symlink_args { false };
    bool one_per_line { false };
    bool list_directories { false };

    os::ArgumentParser parser;
    parser.add_description("List files contained in a directory (defaults to '.', the current directory)"_sv);
    parser.add_system_program_info("ls"_sv);
    parser.add_positional_argument(pathname, "directory"_sv, "."_sv);
    parser.add_switch_argument(show_all, 'a', "all"_sv, "also list hidden files (whose filename begins with a dot)"_sv);
    parser.add_switch_argument(show_almost_all, 'A', "almost-all"_sv, "list all files except '.' and '..'"_sv);
    parser.add_switch_argument(long_list, 'l', ""_sv, "use a long listing format"_sv);
    parser.add_switch_argument(human_readable, 'h', "human-readable"_sv,
                               "with -l, show human-readable sizes e.g. 2KiB, 6GiB"_sv);
    parser.add_switch_argument(si, ' ', "si"_sv, "same as -h, but show sizes in powers of 10"_sv);
    parser.add_switch_argument(follow_symlink_args, 'H', "dereference-args"_sv,
                               "follow symbolic links listed as arguments"_sv);
    parser.add_switch_argument(one_per_line, '1', ""_sv, "list one file per line"_sv);
    parser.add_switch_argument(list_directories, 'd', "directory"_sv, "list directories instead of their contents"_sv);
    parser.parse(argc, argv);

    Vector<String> files;
    int dirfd = AT_FDCWD;
    SharedPtr<os::Directory> dir;

    if (!long_list) follow_symlink_args = true;

    if (os::FileSystem::is_directory(pathname, follow_symlink_args) && !list_directories)
    {
        dir = TRY(os::Directory::open(pathname));
        dirfd = dir->fd();

        auto filter = os::Directory::Filter::Hidden;
        if (show_almost_all) filter = os::Directory::Filter::ParentAndBase;
        else if (show_all)
            filter = os::Directory::Filter::None;

        files = TRY(dir->list(filter));
    }
    else if (os::FileSystem::exists(pathname, follow_symlink_args))
    {
        auto str = TRY(String::from_string_view(pathname));
        TRY(files.try_append(move(str)));
    }
    else
        return err(ENOENT);

    if (!long_list)
    {
        auto list = TRY(String::join(files, one_per_line ? "\n"_sv : " "_sv));
        if (!list.is_empty()) os::println("%s", list.chars());
    }
    else
    {
        for (const auto& file : files)
        {
            struct stat st;
            TRY(os::FileSystem::stat({ dirfd, file.view() }, st, false));

            auto link = TRY(os::FileSystem::readlink({ dirfd, file.view() }));

            StringView owner;
            StringView group;

            find_user_and_group(st, owner, group);

            if (!human_readable && !si)
            {
                os::println("%6o %u %4s %4s %10lu %s%s%s", st.st_mode, st.st_nlink, owner.chars(), group.chars(),
                            st.st_size, file.chars(), link.is_empty() ? "" : " -> ", link.chars());
            }
            else
            {
                auto size = TRY(to_dynamic_unit(st.st_size, 10, false, si ? Unit::SI : Unit::Binary, false));
                os::println("%6o %u %4s %4s %6s %s%s%s", st.st_mode, st.st_nlink, owner.chars(), group.chars(),
                            size.chars(), file.chars(), link.is_empty() ? "" : " -> ", link.chars());
            }
        }
    }

    return 0;
}
